/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2011
 *    Denis Klimentjew <klimentjew@informatik.uni-hamburg.de>
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "pmd_camcube/camcube.h"
#include <sensor_msgs/image_encodings.h>
#include <boost/make_shared.hpp>
#include <cv_bridge/CvBridge.h>

namespace pmd_camcube
{

/** \brief Constructor */
CamcubeDriver::CamcubeDriver (ros::NodeHandle comm_nh, ros::NodeHandle param_nh)

{
  width_ = 200;
  height_ = 200;
  
  // Assemble the point cloud data
  std::string camcube_depth_frame;
  param_nh.param ("camcube_depth_frame", camcube_depth_frame, std::string ("/camcube_depth"));
  cloud_.header.frame_id = cloud2_.header.frame_id = camcube_depth_frame;
  cloud_.channels.resize (1);
  cloud_.channels[0].name = "grayscale";
  cloud_.channels[0].values.resize (width_ * height_);

  cloud2_.height = height_;
  cloud2_.width = width_;
  cloud2_.fields.resize (4);
  cloud2_.fields[0].name = "x";
  cloud2_.fields[1].name = "y";
  cloud2_.fields[2].name = "z";
  cloud2_.fields[3].name = "grayscale";

  // Set all the fields types accordingly
  int offset = 0;
  for (size_t s = 0; s < cloud2_.fields.size (); ++s, offset += 4)
  {
    cloud2_.fields[s].offset   = offset;
    cloud2_.fields[s].count    = 1;
    cloud2_.fields[s].datatype = sensor_msgs::PointField::FLOAT32;
  }

  cloud2_.point_step = offset;
  cloud2_.row_step   = cloud2_.point_step * cloud2_.width;
  cloud2_.data.resize (cloud2_.row_step   * cloud2_.height);

  // Publishers and subscribers
  image_transport::ImageTransport it_(comm_nh);
  sensor_msgs::CvBridge bridge_;
  image_pub_ = it_.advertise("img", 1);
  pub_depth_points_  = comm_nh.advertise<sensor_msgs::PointCloud> ("depth/points", 15);
  pub_depth_points2_ = comm_nh.advertise<sensor_msgs::PointCloud2>("depth/points2", 15);
}


void CamcubeDriver::processGrayscaleAndDepth (PMDHandle hnd){
  
  int res = pmdUpdate (hnd);
  float dat[width_ * height_ * 3];
  unsigned flags[width_ * height_ * 3];

  float *inten = new float[width_*height_];
  int *bet = new int[width_*height_];
  int res2; 
  float max = 0.0f;

  if (pub_depth_points_.getNumSubscribers () > 0 )
  {
    // Assemble an ancient sensor_msgs/PointCloud message
    cloud_.points.resize (0);
    cloud_.channels[0].values.resize(0);
    
    if (res != PMD_OK) {
        pmdGetLastError (0, err, 128);
	ROS_WARN ("Error1: %s\n", err);
	}

    res = pmdGet3DCoordinates(hnd, dat, sizeof(dat));


    
    for(int i = 0; i<(width_ * height_ * 3); i+=3){
      
        // Fill in XYZ
        geometry_msgs::Point32 pt;
	
        pt.x = dat[i];
        pt.y = dat[i+1];
        pt.z = dat[i+2];
        cloud_.points.push_back(pt);

	}
	
    res = pmdGetFlags (hnd, flags, sizeof(flags));

    if (res != PMD_OK) {
        pmdGetLastError (0, err, 128);
        ROS_WARN ("Error3: %s\n", err);
	}
  }
  
  if (pub_depth_points2_.getNumSubscribers () > 0 )
  {
    
    if (res != PMD_OK) {
        pmdGetLastError (0, err, 128);
	ROS_WARN ("Error1: %s\n", err);
	}

    res = pmdGet3DCoordinates(hnd, dat, sizeof(dat));
    
    res2 = pmdGetIntensities(hnd, inten, width_*height_*sizeof(float));

  if (res2 != PMD_OK) {
      pmdGetLastError (0, err, 128);
      ROS_WARN ("Problem2: %s\n", err);
    }

  for(int k = 0; k < width_*height_; k++){
      if(inten[k] > max) max = inten[k];
  }
 
  
  for(int r = 0; r < width_*height_; r++) {
	bet[r] = (int) (255.0 * inten[r] / max);
  }
    
    

    int k = 0;
    for(int m = 0; m < width_*height_*3; m+=3, k++){
        float* pt_data = reinterpret_cast<float*>(&cloud2_.data[0] + k * cloud2_.point_step);
	
        pt_data[0] = dat[m];
        pt_data[1] = dat[m + 1];
        pt_data[2] = dat[m + 2];

        int32_t grayscale_packed = 0;
	grayscale_packed = bet[k];
        memcpy(&pt_data[3], &grayscale_packed, sizeof(int32_t));
    }
	
    res = pmdGetFlags (hnd, flags, sizeof(flags));

    if (res != PMD_OK) {
        pmdGetLastError (0, err, 128);
        ROS_WARN ("Error3: %s\n", err);
	}
	
}

  publish();
}

void CamcubeDriver::publishImg (PMDHandle hnd, IplImage *img){
  
  float *inten = new float[width_*height_]; 
  float max = 0.0f;

  int res = pmdUpdate (hnd);
  
  if (res != PMD_OK) {
      pmdGetLastError (0, err, 128);
      ROS_WARN ("Problem1: %s\n", err);
    }

  res = pmdGetIntensities(hnd, inten, width_*height_*sizeof(float)); 

  if (res != PMD_OK) {
      pmdGetLastError (0, err, 128);
      ROS_WARN ("Problem2: %s\n", err);
    }

  for(int k = 0; k < width_*height_; k++){
      if(inten[k] > max) max = inten[k];
  }
  
  //ROS_INFO("inten %f", max);
  for(int r = 0; r < height_; r++) {
        for(int c = 0; c < width_; c++) {
	    CV_IMAGE_ELEM(img, uint8_t, r, width_ - c) =  (uint8_t) 255.0f * inten[r*height_ + c] / max;
	}    
  }
  
  //ROS_INFO("ok");
  
  if (image_pub_.getNumSubscribers () > 0){
      image_pub_.publish(bridge_.cvToImgMsg(img, "mono8"));
  }
}

void CamcubeDriver::publish (){
  
  /// @todo Do something with the real timestamps from the device
  // Get the current time
  ros::Time time = ros::Time::now ();
  cloud_.header.stamp = cloud2_.header.stamp = time;

  // Publish the PointCloud messages
  if (pub_depth_points_.getNumSubscribers () > 0 )
    pub_depth_points_.publish  (boost::make_shared<const sensor_msgs::PointCloud> (cloud_));
  if (pub_depth_points2_.getNumSubscribers () > 0 )
    pub_depth_points2_.publish (boost::make_shared<const sensor_msgs::PointCloud2> (cloud2_));
}

void CamcubeDriver::showSerialNumber (PMDHandle hnd){
       int res;
	char dat[128];

	res = pmdSourceCommand (hnd, dat, sizeof(dat), "GetSerialNumber");
	if (res != PMD_OK) {
	      pmdGetLastError (0, err, 128);
	      fprintf (stderr, "Error1: %s\n", err);
	}

	ROS_INFO("PMD CamCube Serial Number: %s\n", dat);
}

/** \brief Destructor */
CamcubeDriver::~CamcubeDriver ()
{
    pmdClose (hnd);
}

} // namespace pmd_camcube
